package define

const (
	// TypeDevpts is the type for creating a devpts
	TypeDevpts = "devpts"
	// TypeTmpfs is the type for mounting tmpfs
	TypeTmpfs = "tmpfs"
	// TypeRamfs is the type for mounting ramfs
	TypeRamfs = "ramfs"
	// TypeVolume is the type for named volumes
	TypeVolume = "volume"
)
