// Code generated by go generate; DO NOT EDIT.
package manifests

import (
	"net/url"

	"github.com/containers/podman/v5/pkg/bindings/internal/util"
)

// Changed returns true if named field has been set
func (o *CreateOptions) Changed(fieldName string) bool {
	return util.Changed(o, fieldName)
}

// ToParams formats struct fields to be passed to API service
func (o *CreateOptions) ToParams() (url.Values, error) {
	return util.ToParams(o)
}

// WithAll set field All to given value
func (o *CreateOptions) WithAll(value bool) *CreateOptions {
	o.All = &value
	return o
}

// GetAll returns value of field All
func (o *CreateOptions) GetAll() bool {
	if o.All == nil {
		var z bool
		return z
	}
	return *o.All
}

// WithAmend set field Amend to given value
func (o *CreateOptions) WithAmend(value bool) *CreateOptions {
	o.Amend = &value
	return o
}

// GetAmend returns value of field Amend
func (o *CreateOptions) GetAmend() bool {
	if o.Amend == nil {
		var z bool
		return z
	}
	return *o.Amend
}

// WithAnnotation set field Annotation to given value
func (o *CreateOptions) WithAnnotation(value map[string]string) *CreateOptions {
	o.Annotation = value
	return o
}

// GetAnnotation returns value of field Annotation
func (o *CreateOptions) GetAnnotation() map[string]string {
	if o.Annotation == nil {
		var z map[string]string
		return z
	}
	return o.Annotation
}
