// Code generated by go generate; DO NOT EDIT.
package images

import (
	"net/url"

	"github.com/containers/podman/v5/pkg/bindings/internal/util"
)

// Changed returns true if named field has been set
func (o *TreeOptions) Changed(fieldName string) bool {
	return util.Changed(o, fieldName)
}

// ToParams formats struct fields to be passed to API service
func (o *TreeOptions) ToParams() (url.Values, error) {
	return util.ToParams(o)
}

// WithWhatRequires set field WhatRequires to given value
func (o *TreeOptions) WithWhatRequires(value bool) *TreeOptions {
	o.WhatRequires = &value
	return o
}

// GetWhatRequires returns value of field WhatRequires
func (o *TreeOptions) GetWhatRequires() bool {
	if o.WhatRequires == nil {
		var z bool
		return z
	}
	return *o.WhatRequires
}
