// Code generated by go generate; DO NOT EDIT.
package images

import (
	"net/url"

	"github.com/containers/podman/v5/pkg/bindings/internal/util"
)

// Changed returns true if named field has been set
func (o *ImportOptions) Changed(fieldName string) bool {
	return util.Changed(o, fieldName)
}

// ToParams formats struct fields to be passed to API service
func (o *ImportOptions) ToParams() (url.Values, error) {
	return util.ToParams(o)
}

// WithChanges set field Changes to given value
func (o *ImportOptions) WithChanges(value []string) *ImportOptions {
	o.Changes = &value
	return o
}

// GetChanges returns value of field Changes
func (o *ImportOptions) GetChanges() []string {
	if o.Changes == nil {
		var z []string
		return z
	}
	return *o.Changes
}

// WithMessage set field Message to given value
func (o *ImportOptions) WithMessage(value string) *ImportOptions {
	o.Message = &value
	return o
}

// GetMessage returns value of field Message
func (o *ImportOptions) GetMessage() string {
	if o.Message == nil {
		var z string
		return z
	}
	return *o.Message
}

// WithReference set field Reference to given value
func (o *ImportOptions) WithReference(value string) *ImportOptions {
	o.Reference = &value
	return o
}

// GetReference returns value of field Reference
func (o *ImportOptions) GetReference() string {
	if o.Reference == nil {
		var z string
		return z
	}
	return *o.Reference
}

// WithURL set field URL to given value
func (o *ImportOptions) WithURL(value string) *ImportOptions {
	o.URL = &value
	return o
}

// GetURL returns value of field URL
func (o *ImportOptions) GetURL() string {
	if o.URL == nil {
		var z string
		return z
	}
	return *o.URL
}

// WithOS set field OS to given value
func (o *ImportOptions) WithOS(value string) *ImportOptions {
	o.OS = &value
	return o
}

// GetOS returns value of field OS
func (o *ImportOptions) GetOS() string {
	if o.OS == nil {
		var z string
		return z
	}
	return *o.OS
}

// WithArchitecture set field Architecture to given value
func (o *ImportOptions) WithArchitecture(value string) *ImportOptions {
	o.Architecture = &value
	return o
}

// GetArchitecture returns value of field Architecture
func (o *ImportOptions) GetArchitecture() string {
	if o.Architecture == nil {
		var z string
		return z
	}
	return *o.Architecture
}

// WithVariant set field Variant to given value
func (o *ImportOptions) WithVariant(value string) *ImportOptions {
	o.Variant = &value
	return o
}

// GetVariant returns value of field Variant
func (o *ImportOptions) GetVariant() string {
	if o.Variant == nil {
		var z string
		return z
	}
	return *o.Variant
}
