package network

// This file was generated by the swagger tool.
// Editing this file might prove futile when you re-run the swagger generate command

// CreateResponse NetworkCreateResponse
//
// OK response to NetworkCreate operation
// swagger:model CreateResponse
type CreateResponse struct {

	// The ID of the created network.
	// Required: true
	ID string `json:"Id"`

	// Warnings encountered when creating the container
	// Required: true
	Warning string `json:"Warning"`
}
