// Code generated by go generate; DO NOT EDIT.
package artifacts

import (
	"net/url"

	"github.com/containers/podman/v5/pkg/bindings/internal/util"
)

// Changed returns true if named field has been set
func (o *ExtractOptions) Changed(fieldName string) bool {
	return util.Changed(o, fieldName)
}

// ToParams formats struct fields to be passed to API service
func (o *ExtractOptions) ToParams() (url.Values, error) {
	return util.ToParams(o)
}

// WithTitle set field Title to given value
func (o *ExtractOptions) WithTitle(value string) *ExtractOptions {
	o.Title = &value
	return o
}

// GetTitle returns value of field Title
func (o *ExtractOptions) GetTitle() string {
	if o.Title == nil {
		var z string
		return z
	}
	return *o.Title
}

// WithDigest set field Digest to given value
func (o *ExtractOptions) WithDigest(value string) *ExtractOptions {
	o.Digest = &value
	return o
}

// GetDigest returns value of field Digest
func (o *ExtractOptions) GetDigest() string {
	if o.Digest == nil {
		var z string
		return z
	}
	return *o.Digest
}

// WithExcludeTitle set field ExcludeTitle to given value
func (o *ExtractOptions) WithExcludeTitle(value bool) *ExtractOptions {
	o.ExcludeTitle = &value
	return o
}

// GetExcludeTitle returns value of field ExcludeTitle
func (o *ExtractOptions) GetExcludeTitle() bool {
	if o.ExcludeTitle == nil {
		var z bool
		return z
	}
	return *o.ExcludeTitle
}
