// Code generated by go generate; DO NOT EDIT.
package containers

import (
	"net/url"

	"github.com/containers/podman/v5/pkg/bindings/internal/util"
)

// Changed returns true if named field has been set
func (o *StartOptions) Changed(fieldName string) bool {
	return util.Changed(o, fieldName)
}

// ToParams formats struct fields to be passed to API service
func (o *StartOptions) ToParams() (url.Values, error) {
	return util.ToParams(o)
}

// WithDetachKeys set field DetachKeys to given value
func (o *StartOptions) WithDetachKeys(value string) *StartOptions {
	o.DetachKeys = &value
	return o
}

// GetDetachKeys returns value of field DetachKeys
func (o *StartOptions) GetDetachKeys() string {
	if o.DetachKeys == nil {
		var z string
		return z
	}
	return *o.DetachKeys
}

// WithRecursive set field Recursive to given value
func (o *StartOptions) WithRecursive(value bool) *StartOptions {
	o.Recursive = &value
	return o
}

// GetRecursive returns value of field Recursive
func (o *StartOptions) GetRecursive() bool {
	if o.Recursive == nil {
		var z bool
		return z
	}
	return *o.Recursive
}
